# Copyright (c) Kyutai, all rights reserved.
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import inspect
import os
import secrets
import time
from pathlib import Path
from typing import Optional

import torch
import typer
from aiohttp import web
from moshi.models import LMGen, loaders
from typing_extensions import Annotated

from hibiki_zero.client_utils import audio_read, log, save_results, stack_and_pad_audio
from hibiki_zero.inference import ServerState, decode_outputs, encode_inputs, get_lmgen, seed_all

MODULE_DIR: Path = Path(__file__).parent
DEFAULT_REPO: str = "kyutai/hibiki-zero-3b-pytorch-bf16@23b3e0b41782026c81dd5283a034107b01f9e513"
DEFAULT_AUDIO_SAMPLES: list[Path] = [
    MODULE_DIR / "samples" / fname for fname in os.listdir(MODULE_DIR / "samples")
]
DEFAULT_STATIC_DIR = MODULE_DIR / "static"

cli_app = typer.Typer()


@cli_app.command()
@torch.no_grad()
def serve(
    host: Annotated[str, typer.Option(help="Host to bind the server to.")] = "localhost",
    port: Annotated[int, typer.Option(help="Port to bind the server to.")] = 8998,
    static: Annotated[
        Path, typer.Option(help="Path to static files directory.")
    ] = DEFAULT_STATIC_DIR,
    gradio_tunnel: Annotated[bool, typer.Option(help="Activate a gradio tunnel.")] = False,
    gradio_tunnel_token: Annotated[Optional[str], typer.Option(help="Custom tunnel token.")] = None,
    hf_repo: Annotated[
        str, typer.Option(help="HF repo for model, codec and text tokenizer.")
    ] = DEFAULT_REPO,
    config_path: Annotated[Optional[str], typer.Option(help="Path to a config file.")] = None,
    tokenizer: Annotated[Optional[str], typer.Option(help="Path to a text tokenizer file.")] = None,
    model_weight: Annotated[
        Optional[str], typer.Option(help="Path to a Hibiki-Zero checkpoint.")
    ] = None,
    mimi_weight: Annotated[
        Optional[str], typer.Option(help="Path to a Mimi codec checkpoint.")
    ] = None,
    lora_weight: Annotated[Optional[str], typer.Option(help="Path to a LoRA checkpoint.")] = None,
    fuse_lora: Annotated[
        bool, typer.Option("--fuse-lora/--no-fuse-lora", help="Fuse LoRA layers.")
    ] = True,
    bf16: Annotated[bool, typer.Option(help="Use bfloat16.")] = False,
    device: Annotated[str, typer.Option(help="Device to run on.")] = "cuda",
    ssl: Annotated[
        Optional[str], typer.Option(help="Directory containing cert.pem and key.pem.")
    ] = None,
    seed: Annotated[int, typer.Option(help="Random seed.")] = 42,
):
    # sanity checks
    if not torch.cuda.is_available():
        log(
            "error",
            "Found no NVIDIA driver on your system. The server needs to be launched from a machine that has access to a GPU.",
        )
        return

    seed_all(seed)
    dtype = torch.bfloat16 if bf16 else torch.float16

    if not static.exists():
        log(
            "error",
            "Static frontend directory {0} doesn't exist, please run {1} first.",
            [(static, "grey"), (MODULE_DIR.parent / "build_frontend.sh", "orange")],
        )
        return

    log("info", "Starting Hibiki-Zero server.")
    setup_tunnel, tunnel_token = None, ""
    if gradio_tunnel:
        try:
            from gradio import networking  # type: ignore
        except ImportError:
            log(
                "error", "Gradio is required for tunnel support. Install with `pip install gradio`."
            )
            raise typer.Exit(1)
        setup_tunnel = networking.setup_tunnel
        tunnel_token = (
            secrets.token_urlsafe(32) if gradio_tunnel_token is None else gradio_tunnel_token
        )

    hf_repo_parts = hf_repo.split("@")
    hf_repo_name = hf_repo_parts[0]
    if len(hf_repo_parts) > 1:
        revision = hf_repo_parts[1]
    else:
        revision = None

    log("info", "Retrieving the model checkpoint...")
    checkpoint_info = loaders.CheckpointInfo.from_hf_repo(
        hf_repo_name,
        model_weight,
        mimi_weight,
        tokenizer,
        lora_weights=lora_weight,
        config_path=config_path,
        revision=revision,
    )

    log("info", "Loading the codec {0}", [(checkpoint_info.mimi_weights, "blue")])
    mimi = checkpoint_info.get_mimi(device=device)
    text_tokenizer = checkpoint_info.get_text_tokenizer()

    log("info", "Loading the model {0}", [(checkpoint_info.moshi_weights, "blue")])
    lm = checkpoint_info.get_moshi(device=device, dtype=dtype, fuse_lora=fuse_lora)

    state = ServerState(
        checkpoint_info.model_type,
        mimi,
        text_tokenizer,
        lm,
        device,
        **checkpoint_info.lm_gen_config,
    )
    log("info", "Warming up the model...")
    state.warmup()

    web_app = web.Application()
    web_app.router.add_get("/api/chat", state.handle_chat)

    async def handle_root(_):
        return web.FileResponse(os.path.join(static, "index.html"))

    web_app.router.add_get("/", handle_root)
    web_app.router.add_static("/", path=static, follow_symlinks=True, name="static")

    protocol, ssl_context = "http", None
    if ssl is not None:
        import ssl as ssl_module

        ssl_context = ssl_module.create_default_context(ssl_module.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(
            certfile=os.path.join(ssl, "cert.pem"), keyfile=os.path.join(ssl, "key.pem")
        )
        protocol = "https"

    local_url = f"{protocol}://{host}:{port}"
    log("info", "Access the Web UI directly at {0}", [(local_url, "orange")])

    if setup_tunnel is not None:
        tunnel_kwargs = {}
        if "share_server_tls_certificate" in inspect.signature(setup_tunnel).parameters:
            tunnel_kwargs["share_server_tls_certificate"] = None
        tunnel = setup_tunnel("localhost", port, tunnel_token, None, **tunnel_kwargs)  # type: ignore
        log("info", "Tunnel started at {0}", [(tunnel, "green")])
        log("info", "Note: tunnel goes through the US; expect higher latency in Europe.")

    web.run_app(web_app, host=host, port=port, ssl_context=ssl_context)


@cli_app.command()
@torch.no_grad()
def generate(
    files: Annotated[list[Path], typer.Option("--file", help="Input files to translate.")] = None,
    gen_duration: Annotated[
        float,
        typer.Option(
            help="Generation duration in seconds. Should be <=120 seconds for Hibiki-Zero."
        ),
    ] = 120,
    out_dir: Annotated[Path, typer.Option(help="Directory where to save the outputs.")] = Path(
        "./translations"
    ),
    tag: Annotated[
        str, typer.Option(help="Tag to add to translation outputs filenames to identify them.")
    ] = None,
    repeats: Annotated[int, typer.Option(help="Do repeats generation for each input file.")] = 1,
    hf_repo: Annotated[
        str, typer.Option(help="HF repo for model, codec and text tokenizer.")
    ] = DEFAULT_REPO,
    config_path: Annotated[Optional[str], typer.Option(help="Path to a config file.")] = None,
    tokenizer: Annotated[Optional[str], typer.Option(help="Path to a text tokenizer file.")] = None,
    model_weight: Annotated[
        Optional[str], typer.Option(help="Path to a Hibiki-Zero checkpoint.")
    ] = None,
    mimi_weight: Annotated[
        Optional[str], typer.Option(help="Path to a Mimi codec checkpoint.")
    ] = None,
    lora_weight: Annotated[Optional[str], typer.Option(help="Path to a LoRA checkpoint.")] = None,
    fuse_lora: Annotated[
        bool, typer.Option("--fuse-lora/--no-fuse-lora", help="Fuse LoRA layers.")
    ] = True,
    bf16: Annotated[bool, typer.Option(help="Use bfloat16.")] = False,
    device: Annotated[str, typer.Option(help="Device to run on.")] = "cuda",
    seed: Annotated[int, typer.Option(help="Random seed.")] = 42,
):
    if not torch.cuda.is_available():
        log(
            "error",
            "Found no NVIDIA driver on your system. Generation needs to be done on a machine that has access to a GPU.",
        )
        return

    seed_all(seed)
    dtype = torch.bfloat16 if bf16 else torch.float16

    log("info", "Starting Hibiki-Zero inference.")
    files = files if files is not None else DEFAULT_AUDIO_SAMPLES
    files = [fpath for fpath in files for _ in range(repeats)]
    all_files_exist: bool = len(files) > 0
    for fpath in files:
        if not fpath.exists():
            log("error", f"File not found: {fpath}")
            all_files_exist = False
    if not all_files_exist:
        if len(files) == 0:
            log("error", "No files provided.")
        return
    log("info", "The following audios will be processed in a single batch:")
    for fidx, fpath in enumerate(files):
        log("info", f"{fidx} : " + "{0}", [(fpath, "grey")])

    hf_repo_parts = hf_repo.split("@")
    hf_repo_name = hf_repo_parts[0]
    if len(hf_repo_parts) > 1:
        revision = hf_repo_parts[1]
    else:
        revision = None

    log("info", "Retrieving the model checkpoint...")
    checkpoint_info = loaders.CheckpointInfo.from_hf_repo(
        hf_repo_name,
        model_weight,
        mimi_weight,
        tokenizer,
        lora_weights=lora_weight,
        config_path=config_path,
        revision=revision,
    )

    log("info", "Loading the codec {0}", [(checkpoint_info.mimi_weights, "blue")])
    mimi = checkpoint_info.get_mimi(device=device)
    text_tokenizer = checkpoint_info.get_text_tokenizer()

    log("info", "Loading the model {0}", [(checkpoint_info.moshi_weights, "blue")])
    lm = checkpoint_info.get_moshi(device=device, dtype=dtype, fuse_lora=fuse_lora)

    log("info", "Loading audios...")
    input_wavs: list[torch.Tensor] = [
        audio_read(fpath, to_sample_rate=mimi.sample_rate, mono=True)[0] for fpath in files
    ]
    audio_durations: list[float] = [wav.shape[-1] / mimi.sample_rate for wav in input_wavs]
    if max(audio_durations) > gen_duration:
        log(
            "error",
            f"One of the input audios is longer than the gen duration: {max(audio_durations)} > {gen_duration=}",
        )
        return
    batch_wavs = stack_and_pad_audio(input_wavs, max_len=int(gen_duration * mimi.sample_rate))
    batch_size: int = batch_wavs.shape[0]

    lm_gen: LMGen = get_lmgen(lm, checkpoint_info, batch_size)

    log("info", "Encoding audios...")
    codes, warmup_codes = encode_inputs(batch_wavs, mimi, lm_gen, audio_durations)

    output_text_tokens: list[torch.Tensor] = []
    output_audio_tokens: list[torch.Tensor] = []
    gen_steps: int = codes.shape[-1]
    start_gen_time: float = time.time()
    with torch.no_grad(), lm_gen.streaming(batch_size):
        # warmup
        for step in range(warmup_codes.shape[-1]):
            _ = lm_gen.step(warmup_codes[:, :, step : step + 1])
        # generation
        for step in range(codes.shape[-1]):
            tokens = lm_gen.step(codes[:, :, step : step + 1])
            if tokens is None:
                print(None)
            else:
                output_text_tokens.append(tokens[:, 0, :])
                output_audio_tokens.append(tokens[:, 1:, :])
            log(
                "info",
                f"Running inference: {step}/{gen_steps} steps = {step / gen_steps:.0%}",
                end="\r",
            )
    gen_time: float = time.time() - start_gen_time
    real_time_factor: float = batch_wavs.shape[-1] / mimi.sample_rate / gen_time
    throughput: float = real_time_factor * batch_size
    log(
        "info",
        "Generated outputs in {0} "
        + f"(throughput = batch size x real-time factor = {throughput:.1f})",
        [(f"{real_time_factor:.1f}x real-time", "orange")],
    )

    log("info", "Saving results...")
    batch_text_tokens: torch.Tensor = torch.concat(output_text_tokens, dim=-1)  # B x T
    batch_codes: torch.Tensor = torch.concat(output_audio_tokens, dim=-1)  # B x K x T
    outputs = decode_outputs(batch_codes, batch_text_tokens, mimi, text_tokenizer)
    save_results(
        inputs=zip(files, input_wavs),
        outputs=outputs,
        sample_rate=mimi.sample_rate,
        output_dir=out_dir,
        tag=tag,
    )
    log("info", "Saved translation results in {0}", [(out_dir, "green")])


def main():
    cli_app()


if __name__ == "__main__":
    main()



# Copyright (c) Kyutai, all rights reserved.
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

import sphn
import torch
from IPython.display import Audio, display

COLOR_MAP: dict[str, tuple] = {
    "blue": (92, 158, 255),
    "yellow": (255, 255, 85),
    "red": (255, 85, 85),
    "orange": (255, 171, 64),
    "green": (57, 242, 174),
    "grey": (128, 128, 128),
}


def get_color_code(color_name: str) -> tuple:
    return COLOR_MAP.get(color_name, (255, 255, 255))  # default to white if unknown


def colorize_rgb(text: str, rgb: tuple) -> str:
    return f"\033[38;2;{rgb[0]};{rgb[1]};{rgb[2]}m{text}\033[0m"


def make_colored_log(
    level: str, msg: str, colored_parts: list[tuple[str, str]] | None = None
) -> str:
    if level == "info":
        prefix = colorize_rgb("[Info]", get_color_code("blue"))
    elif level == "warning":
        prefix = colorize_rgb("[Warn]", get_color_code("yellow"))
    elif level == "error":
        prefix = colorize_rgb("[Err ]", get_color_code("red"))
    else:
        raise ValueError(f"Unknown level {level}")
    if colored_parts is not None:
        msg = msg.format(
            *[colorize_rgb(text, get_color_code(color_code)) for text, color_code in colored_parts]
        )
    return prefix + " " + msg


def log(
    level: str, msg: str, colored_parts: list[tuple[str, str]] | None = None, end: str | None = "\n"
) -> None:
    """Log something with a given level."""
    print(make_colored_log(level, msg, colored_parts), end=end)


def audio_read(
    fpath: Path, to_sample_rate: int | None = None, mono: bool = False
) -> tuple[torch.Tensor, int]:
    """Read audio fpath and resample at to_sample_rate/transform to mono audio if specified."""
    wav, sr = sphn.read(fpath)
    if to_sample_rate is not None and sr != to_sample_rate:
        wav = sphn.resample(wav, sr, to_sample_rate)
        sr = to_sample_rate
    wav_tensor: torch.Tensor = torch.tensor(wav)
    if wav_tensor.ndim == 1:
        wav_tensor.unsqueeze(0)
    elif wav_tensor.ndim == 2:
        if wav_tensor.shape[0] > 2:
            raise ValueError(
                f"Audio {fpath} has too many channels, got {wav_tensor.shape[0]} but expected 1 or 2."
            )
        elif wav_tensor.shape[0] == 2 and mono:
            log(
                "info",
                "Audio {0} is stereo, averaging both channels to get a mono audio.",
                [(fpath, "grey")],
            )
            wav_tensor = wav_tensor.mean(dim=0, keepdim=True)
    elif wav_tensor.ndim >= 3:
        raise ValueError(
            f"Audio {fpath} was loaded into a tensor of unsupported shape {wav_tensor.ndim}"
        )
    return wav_tensor, sr


def stack_and_pad_audio(wavs: list[torch.Tensor], max_len: int | None = None) -> torch.Tensor:
    """Stack the given audios on the first dimenion (created), padding them with 0 if needed."""
    actual_max_len = max(wav.shape[-1] for wav in wavs)
    if max_len is None:
        max_len = actual_max_len
    else:
        assert actual_max_len <= max_len, (actual_max_len, max_len)
    other_dims = wavs[0].shape[:-1]
    out = torch.zeros(len(wavs), *other_dims, max_len, dtype=wavs[0].dtype, device=wavs[0].device)
    for k, wav in enumerate(wavs):
        out[k, ..., : wav.shape[-1]] = wav
    return out


def display_audio(wav: torch.Tensor, sample_rate: int):
    """Display an audio reader to be used in an interactive notebook."""
    display(Audio(wav.numpy(), rate=sample_rate))


def save_results(
    inputs: list[str, torch.Tensor],
    outputs: list[tuple[torch.Tensor, str]],
    sample_rate: int,
    output_dir: Path,
    tag: str | None,
):
    tag_suffix: str = "" if tag is None else f"_{tag}"
    for file_idx, ((in_fpath, in_wav), (out_wav, out_text)) in enumerate(zip(inputs, outputs)):
        stereo_audio: torch.Tensor = stack_and_pad_audio([in_wav, out_wav]).squeeze()
        output_stem_fpath: Path = output_dir / f"{file_idx}_{in_fpath.stem}{tag_suffix}"
        mono_fpath: Path = output_stem_fpath.parent / f"{output_stem_fpath.stem}_mono.wav"
        stereo_fpath: Path = output_stem_fpath.parent / f"{output_stem_fpath.stem}_stereo.wav"
        text_fpath: Path = output_stem_fpath.parent / f"{output_stem_fpath.stem}.txt"
        mono_fpath.parent.mkdir(exist_ok=True, parents=True)
        sphn.write_wav(mono_fpath, out_wav.numpy(), sample_rate)
        sphn.write_wav(stereo_fpath, stereo_audio.numpy(), sample_rate)
        with open(text_fpath, "w") as f:
            f.write(out_text)



# Copyright (c) Kyutai, all rights reserved.
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import asyncio
import math
import random
import time
from dataclasses import dataclass

import aiohttp
import numpy as np
import sentencepiece
import sphn
import torch
from aiohttp import web
from moshi.models import LMGen, LMModel, MimiModel, loaders
from moshi.run_inference import get_condition_tensors

from hibiki_zero.client_utils import log


def seed_all(seed):
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)  # for multi-GPU setups
    random.seed(seed)
    np.random.seed(seed)
    torch.backends.cudnn.deterministic = False
    torch.backends.cudnn.benchmark = False


@dataclass
class ServerState:
    model_type: str
    mimi: MimiModel
    text_tokenizer: sentencepiece.SentencePieceProcessor
    lm_gen: LMGen
    lock: asyncio.Lock

    def __init__(
        self,
        model_type: str,
        mimi: MimiModel,
        text_tokenizer: sentencepiece.SentencePieceProcessor,
        lm: LMModel,
        device: str | torch.device,
        **kwargs,
    ):
        self.model_type = model_type
        self.mimi = mimi
        self.text_tokenizer = text_tokenizer
        condition_tensors = get_condition_tensors(model_type, lm, batch_size=1, cfg_coef=1)
        self.lm_gen = LMGen(lm, cfg_coef=1, condition_tensors=condition_tensors, **kwargs)

        self.device = device
        self.frame_size = int(self.mimi.sample_rate / self.mimi.frame_rate)
        self.lock = asyncio.Lock()

        self.mimi.streaming_forever(1)
        self.lm_gen.streaming_forever(1)

    def warmup(self):
        for chunk in range(4):
            chunk = torch.zeros(1, 1, self.frame_size, dtype=torch.float32, device=self.device)
            codes = self.mimi.encode(chunk)
            for c in range(codes.shape[-1]):
                tokens = self.lm_gen.step(codes[:, :, c : c + 1])
                if tokens is None:
                    continue
                _ = self.mimi.decode(tokens[:, 1:])

        torch.cuda.synchronize()

    async def decode_and_send(
        self, tokens: torch.Tensor, ws: web.WebSocketResponse, opus_writer: sphn.OpusStreamWriter
    ):
        assert tokens.shape[1] == self.lm_gen.lm_model.dep_q + 1
        main_pcm = self.mimi.decode(tokens[:, 1:])
        main_pcm = main_pcm.cpu()
        opus_bytes = opus_writer.append_pcm(main_pcm[0, 0].numpy())
        if len(opus_bytes) > 0:
            await ws.send_bytes(b"\x01" + opus_bytes)
        text_token = tokens[0, 0, 0].item()
        if text_token not in (0, 3):
            _text = self.text_tokenizer.id_to_piece(text_token)  # type: ignore
            _text = _text.replace("▁", " ")
            msg = b"\x02" + bytes(_text, encoding="utf8")
            log("info", f"text token: '{_text}'")
            await ws.send_bytes(msg)
        elif text_token == 2:
            log("info", "End Of Sequence token")

    async def recv_loop(
        self,
        ws: web.WebSocketResponse,
        opus_reader: sphn.OpusStreamReader,
        opus_writer: sphn.OpusStreamWriter,
    ):
        all_pcm_data = None
        skip_frames = 1
        try:
            frame_idx: int = 0
            async for message in ws:
                if message.type == aiohttp.WSMsgType.ERROR:
                    log("error", f"{ws.exception()}")
                    break
                elif message.type == aiohttp.WSMsgType.CLOSED:
                    break
                elif message.type != aiohttp.WSMsgType.BINARY:
                    log("error", f"Unexpected message type {message.type}")
                    continue
                message = message.data
                if not isinstance(message, bytes):
                    log("error", f"Unsupported message type {type(message)}")
                    continue
                if len(message) == 0:
                    log("warning", "Empty message")
                    continue
                kind = message[0]
                if kind == 1:  # audio
                    payload = message[1:]
                    pcm = opus_reader.append_bytes(payload)
                    if pcm.shape[-1] == 0:
                        continue
                    if all_pcm_data is None:
                        all_pcm_data = pcm
                    else:
                        all_pcm_data = np.concatenate((all_pcm_data, pcm))
                    while all_pcm_data.shape[-1] >= self.frame_size:
                        be = time.time()
                        chunk = all_pcm_data[: self.frame_size]
                        all_pcm_data = all_pcm_data[self.frame_size :]
                        chunk = torch.from_numpy(chunk)
                        chunk = chunk.to(device=self.device)[None, None]
                        codes = self.mimi.encode(chunk)
                        if skip_frames:
                            # The first input audio frame is ignored, as from the point of
                            # view of the model it is in the past. We still `mimi.encode` for simplicity,
                            # however as the first encoded frame has a specific structure (due to the left padding),
                            # we reset the streaming state of the encoder to reapply the padding on the next call.
                            self.mimi.reset_streaming()
                            skip_frames -= 1
                        for c in range(codes.shape[-1]):
                            tokens = self.lm_gen.step(codes[:, :, c : c + 1])
                            if tokens is None:
                                continue
                            await self.decode_and_send(tokens, ws, opus_writer)
                        log(
                            "info",
                            f"frame {frame_idx} handled in {1000 * (time.time() - be):.1f}ms",
                        )
                        frame_idx += 1
                else:
                    log("warning", f"unknown message kind {kind}")
        finally:
            log("info", "Connection closed.")

    async def handle_chat(self, request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        log("info", "Accepted connection.")

        async with self.lock:
            opus_writer = sphn.OpusStreamWriter(self.mimi.sample_rate)
            opus_reader = sphn.OpusStreamReader(self.mimi.sample_rate)
            self.mimi.reset_streaming()
            self.lm_gen.reset_streaming()
            # Send the handshake.
            await ws.send_bytes(b"\x00")
            await self.recv_loop(ws, opus_reader, opus_writer)
        log("info", "Done with connection.")
        return ws


def get_lmgen(
    lm: LMModel, checkpoint_info: loaders.CheckpointInfo, batch_size: int, cfg_coef: int = 1.0
) -> LMGen:
    condition_tensors = get_condition_tensors(
        checkpoint_info.model_type, lm, batch_size=batch_size, cfg_coef=cfg_coef
    )
    lm_gen = LMGen(
        lm, cfg_coef=cfg_coef, condition_tensors=condition_tensors, **checkpoint_info.lm_gen_config
    )
    return lm_gen


def add_input_eos(
    codes: torch.Tensor, mimi: MimiModel, audio_durations: list[float]
) -> torch.Tensor:
    other_audio_eos_idx: torch.Tensor = torch.tensor(
        [
            min(math.ceil(duration * mimi.frame_rate), codes.shape[-1] - 1)
            for duration in audio_durations
        ]
    )[:, None, None].to(codes.device)  # B, 1, 1
    codes_like_indexes: torch.Tensor = torch.arange(0, codes.shape[-1])[None, None].to(
        codes.device
    )  # 1, 1, T
    codes_with_input_eos: torch.Tensor = torch.where(
        codes_like_indexes == other_audio_eos_idx,
        torch.full([1], mimi.cardinality, device=codes.device),
        codes,
    )  # B, K, T
    return codes_with_input_eos


def encode_inputs(
    batch_wavs: torch.Tensor, mimi: MimiModel, lm_gen: LMGen, audio_durations: list[float]
) -> tuple[torch.Tensor, torch.Tensor]:
    frame_size = int(mimi.sample_rate / mimi.frame_rate)
    with torch.no_grad():
        codes: torch.Tensor = mimi.encode(batch_wavs.to(lm_gen.lm_model.device))
        codes_with_input_eos = add_input_eos(codes, mimi, audio_durations)
        warmup_wav: torch.Tensor = torch.zeros(
            codes.shape[0],
            1,
            frame_size * lm_gen.max_delay,
            dtype=torch.float32,
            device=codes.device,
        )
        warmup_codes: torch.Tensor = mimi.encode(warmup_wav)
    return codes_with_input_eos, warmup_codes


def decode_outputs(
    batch_codes: torch.Tensor,
    batch_text_tokens: torch.Tensor,
    mimi: MimiModel,
    text_tokenizer: sentencepiece.SentencePieceProcessor,
) -> list[tuple[torch.Tensor, str]]:
    with torch.no_grad():
        output_wavs: torch.Tensor = mimi.decode(batch_codes).cpu()

    outputs: list[tuple[torch.Tensor, str]] = []
    for output_idx, wav in enumerate(output_wavs):
        text_tokens: list[int] = batch_text_tokens[output_idx].tolist()
        if text_tokenizer.eos_id() in text_tokens:
            eos_idx: int = text_tokens.index(text_tokenizer.eos_id())
        else:
            log(
                "warning",
                f"The model didn't generate output EOS token for entry {output_idx}, truncating audio after the last word generated.",
            )
            eos_idx: int = len(text_tokens) - 1
            while eos_idx > 0 and text_tokens[eos_idx] == text_tokenizer.pad_id():
                eos_idx -= 1
        text_tokens = [t for t in text_tokens[:eos_idx] if t > text_tokenizer.pad_id()]
        text: str = text_tokenizer.decode(text_tokens)
        wav = wav[:, : int(eos_idx * mimi.sample_rate / mimi.frame_rate)]
        outputs.append((wav, text))

    return outputs
